/*
 * Copyright © 2006-2009 Spotify AB
 */

/* Must be at the top to have feature macros set */
#include <ngx_config.h>
#include <ngx_sha1.h>
#include "ngx_http_upstream_spdht_module.h"


/** Maximum number of rehashes to find slaves. */
#define MAX_REHASH 64


/** Inplace sha1 of sha1. */
static ngx_int_t
rehash(unsigned char *hash, ngx_log_t *log)
{
	ngx_sha1_t ctx;

	if (!ngx_sha1_init(&ctx) || !ngx_sha1_update(&ctx, hash, 20) || !ngx_sha1_final(hash, &ctx)) {
		ngx_log_error(NGX_LOG_EMERG, log, 0, "DHT: rehash failed!");
		return NGX_ERROR;
	}

	return NGX_OK;
}


/** Fill dest with n master and slave servers. */
ngx_int_t
ngx_http_upstream_spdht_master_slaves(ngx_http_upstream_spdht_domain_t *domain, ngx_str_t master_hash, ngx_int_t n, ngx_array_t *dest /* (ngx_http_upstream_spdht_server_t *) */, ngx_log_t *log)
{
	unsigned char hash[20];
	ngx_http_upstream_spdht_token_t hash_token;
	ngx_uint_t nelts;
	ngx_http_upstream_spdht_token_t *elts;
	ngx_http_upstream_spdht_server_t *server;
	ngx_uint_t i, k;

	if (master_hash.len != 20) {
		ngx_log_error(NGX_LOG_EMERG, log, 0, "DHT: invalid master hash!");
		return NGX_ERROR;
	}
	/* Construct a token from the master hash */
	memcpy(hash, master_hash.data, 20);
	hash_token.token.data = hash;
	hash_token.token.len = 20;
	hash_token.server = NULL;

	nelts = domain->state.tokens.nelts;
	if (nelts == 0 || n <= 0) {
		return NGX_ERROR;
	}
	elts = (ngx_http_upstream_spdht_token_t *) domain->state.tokens.elts;

	for (k = 0; k < MAX_REHASH; ++k) {
		/* Find dht server */
		ngx_int_t node = ngx_http_upstream_spdht_bisect_left(&hash_token, elts, nelts, sizeof(ngx_http_upstream_spdht_token_t), ngx_http_upstream_spdht_token_cmp);
		server = ((ngx_http_upstream_spdht_token_t *) elts)[node].server;

		ngx_int_t dup_host = 0;
		/* Look for duplicate servers in dest */
		for (i = 0; i < dest->nelts; ++i) {
			ngx_http_upstream_spdht_server_t *prev = ((ngx_http_upstream_spdht_server_t **) dest->elts)[i];
			if (server == prev ||
					(server->uri.host.len == prev->uri.host.len &&
					 !ngx_strncmp(server->uri.host.data, prev->uri.host.data, server->uri.host.len))) {
				dup_host = 1;
				break; /* host is unique in dest, break here */
			}
		}
		/* Append to dest if host is unique */
		if (dup_host == 0) {
			ngx_http_upstream_spdht_server_t **p = (ngx_http_upstream_spdht_server_t **) ngx_array_push(dest);
			if (p == NULL) {
				return NGX_ERROR;
			}
			*p = server;
			if ((ngx_int_t) dest->nelts >= n) {
				/* Found enough */
				return NGX_OK;
			}
		}
		/* Rehash the token */
		if (rehash(hash, log) != NGX_OK) {
			return NGX_ERROR;
		}
	}
	/* Found too few, still ok */
	return NGX_OK;
}

