/*
 * Copyright © 2006-2009 Spotify AB
 */

/* Must be at the top to have feature macros set */
#include <ctype.h>
#include <ngx_config.h>
#include <resolv.h>
#include <arpa/nameser.h>
#include "ngx_http_upstream_spdht_module.h"



static ngx_int_t lookup_tokens(ngx_http_upstream_spdht_server_t *server, ngx_array_t *tokens, ngx_log_t *log)
{
	unsigned char ans[4096];
	int anslen;
	ns_msg msg;
	unsigned int nans;
	unsigned int i;
	u_int32_t min_ttl = UINT_MAX;
	u_char txt_name[NS_MAXDNAME];

	/* ngx_snprintf() does not append the zero character */
	ngx_snprintf(txt_name, sizeof(txt_name), "tokens.%ud.%V", server->uri.port, &server->uri.host)[0] = '\0';

	ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "res_query(txt, %s)", txt_name);

	anslen = res_query((const char*) txt_name, ns_c_in, ns_t_txt, ans, sizeof(ans));

	if (anslen < 0) {
		ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to look up TXT RR: %d (TXT record for %s)", h_errno, txt_name);
		return NGX_ERROR;
	}

	if (ns_initparse(ans, anslen, &msg)) {
		ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse TXT answer: %d (TXT record for %s)", h_errno, txt_name);
		return NGX_ERROR;
	}

	nans = ns_msg_count(msg, ns_s_an);

	for (i = 0; i < nans; ++i) {
		ns_rr rr;

		if (ns_parserr(&msg, ns_s_an, i, &rr)) {
			ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse RR in TXT answer: %d", h_errno);
			return NGX_ERROR;
		}

		if (ns_rr_type(rr) != ns_t_txt || ns_rr_class(rr) != ns_c_in)
			continue;

		if (ns_rr_ttl(rr) < min_ttl)
			min_ttl = ns_rr_ttl(rr);

		{
			const u_char *data = ns_rr_rdata(rr);
			ngx_uint_t remaining = ns_rr_rdlen(rr);

			while (remaining) {
				if (data[0] > remaining - 1) {
					ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse TXT rdata");
					return NGX_ERROR;
				}

				if (data[0] == 40) {
					ngx_http_upstream_spdht_token_t *token = (ngx_http_upstream_spdht_token_t *) ngx_array_push(tokens);
					ngx_str_t tok_value = { data[0], (u_char*) data + 1 };

					if (ngx_http_upstream_spdht_parse_hex_string(tokens->pool, &token->token, &tok_value) != NGX_OK) {
						ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse TXT rdata, skipping");
					}

					token->server = server;
				} else {
					ngx_log_error(NGX_LOG_ERR, log, 0, "TXT string invalid length: %ud, skipping", data[0]);
				}

				remaining -= data[0] + 1;
				data += data[0] + 1;
			}
		}

	}

	if (min_ttl != UINT_MAX)
		server->tokens_tod = ngx_time() + min_ttl;

	ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "TXT %s -> %ud records, TTL %ud", txt_name, nans, min_ttl);

	return NGX_OK;
}

char * ngx_http_upstream_spdht_dns_init(ngx_conf_t *cf)
{
	if (res_init()) {
		ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "failed to initialize the resolver library");

		return NGX_CONF_ERROR;
	}

	return NGX_CONF_OK;
}


static
ngx_int_t parse_slaves(ngx_http_upstream_spdht_domain_t *domain, ngx_str_t slaves, ngx_log_t *log) {
	if (ngx_array_init(&domain->state.collections, domain->state.servers.pool, 3, sizeof(ngx_http_upstream_spdht_collection_t)) != NGX_OK) {
		return NGX_ERROR;
	}
	size_t off = 0;
	while (off < slaves.len) {
		/* Skip whitespace */
		while (slaves.data[off] == ' ') {
			if (!(++off < slaves.len)) {
				return NGX_OK;
			}
		}
		/* Search for name:number */
		size_t start = off, colon = off;
		ngx_int_t n = 0;
		while (slaves.data[off] != ' ') {
			if (slaves.data[off] == ':') {
				colon = off;
			}
			else if (colon > start) {
				if (!isdigit(slaves.data[off])) {
					ngx_log_error(NGX_LOG_ERR, log, 0, "NON-DIGIT at %uz in %V", off, &slaves);
					return NGX_ERROR;
				}
				n = n * 10 + slaves.data[off] - '0';
			}
			if (!(++off < slaves.len)) {
				break;
			}
		}
		if (colon == start) {
			ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "DEFAULT :%i", n);
			domain->state.num_slaves = n;
		}
		else {
			ngx_http_upstream_spdht_collection_t *collection = (ngx_http_upstream_spdht_collection_t *) ngx_array_push(&domain->state.collections);
			if (collection == NULL) {
				return NGX_ERROR;
			}
			ngx_str_t name = { len: colon - start, data: slaves.data + start };
			ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "SLAVE %V:%i", &name, n);
			collection->name.data = ngx_pstrdup(domain->state.collections.pool, &name);
			collection->name.len = name.len;
			collection->num_slaves = n;
		}
	}
	return NGX_OK;
}


ngx_int_t ngx_http_upstream_spdht_validate_domain_state(ngx_str_t *srv_prefix, ngx_http_upstream_spdht_domain_t *domain, ngx_str_t *server_name, ngx_log_t *log)
{
	unsigned char ans[4096];
	int anslen;
	ns_msg msg;
	unsigned int nans;
	unsigned int i;
	u_int32_t min_ttl = UINT_MAX;
	u_char srv_name[NS_MAXDNAME];
	u_char txt_name[NS_MAXDNAME];
	ngx_array_t new_servers;
	ngx_array_t new_tokens;

	if (domain->state.tod >= ngx_time())
		return NGX_OK;

	/* ngx_snprintf() does not append the zero character */
	ngx_snprintf(srv_name, sizeof(srv_name), "%V.%V", srv_prefix, &domain->name)[0] = '\0';

	ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "res_query(srv, %s)", srv_name);

	anslen = res_query((const char*) srv_name, ns_c_in, ns_t_srv, ans, sizeof(ans));

	if (anslen < 0) {
		ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to look up SRV RR: %d", h_errno);
		return NGX_ERROR;
	}

	if (ns_initparse(ans, anslen, &msg)) {
		ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse SRV answer: %d", h_errno);
		return NGX_ERROR;
	}

	nans = ns_msg_count(msg, ns_s_an);

	if (ngx_array_init(&new_servers, domain->state.servers.pool, nans, sizeof(ngx_http_upstream_spdht_server_t)) != NGX_OK)
		return NGX_ERROR;

	for (i = 0; i < nans; ++i) {
		ns_rr rr;

		if (ns_parserr(&msg, ns_s_an, i, &rr)) {
			ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse RR in SRV answer: %d", h_errno);
			ngx_array_destroy(&new_servers);
			return NGX_ERROR;
		}

		if (ns_rr_type(rr) != ns_t_srv || ns_rr_class(rr) != ns_c_in)
			continue;

		if (ns_rr_ttl(rr) < min_ttl)
			min_ttl = ns_rr_ttl(rr);

		{
			// The following is in compliance with RFC 2782
			const unsigned char *data = ns_rr_rdata(rr);
			unsigned int prio = ns_get16(data);
			unsigned int weight = ns_get16(data + 2);
			unsigned int port = ns_get16(data + 4);
			char target[NS_MAXDNAME];
			unsigned int j;

			if (ns_name_uncompress(ans, ans + anslen, data + 6, target, sizeof(target)) < 0) {
				ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to uncompress target in SRV RR: %d", h_errno);
				ngx_array_destroy(&new_servers);
				return NGX_ERROR;
			}

			if (!strcmp(target, "."))
				break;

			ngx_log_debug4(NGX_LOG_DEBUG_HTTP, log, 0, "SRV %ud %ud %ud %s", prio, weight, port, target);

			for (j = 0; j < domain->state.servers.nelts; ++j) {
				ngx_http_upstream_spdht_server_t *old_server = ((ngx_http_upstream_spdht_server_t *) domain->state.servers.elts) + j;

				if (!ngx_strncmp(target, old_server->uri.host.data, old_server->uri.host.len) && port == old_server->uri.port) {
					/* It's already in here, just copy it. */
					ngx_http_upstream_spdht_server_t *new_server = (ngx_http_upstream_spdht_server_t *) ngx_array_push(&new_servers);

					memcpy(new_server, old_server, sizeof(*old_server));
					old_server->new_server = new_servers.nelts - 1; /* Used by the token copying loop below */
					if ((int) weight != old_server->weight) {
						/* Rescale current_weight and update weight */
						if (old_server->weight > 0)
							new_server->current_weight = old_server->current_weight * weight / old_server->weight;
						else
							new_server->current_weight = 0;
						new_server->weight = weight;
					}
					break;
				}
			}

			if (j == domain->state.servers.nelts) {
				ngx_http_upstream_spdht_server_t *server = (ngx_http_upstream_spdht_server_t *) ngx_array_push(&new_servers);
				ngx_str_t target_str = { strlen(target), (u_char*) target };

				server->tokens_tod = -1;
				ngx_memset(&server->uri, 0, sizeof(server->uri));
				server->uri.host.data = ngx_pstrdup(new_servers.pool, &target_str);
				server->uri.host.len = target_str.len;
				server->uri.port = server->uri.default_port = port;
				server->uri.family = AF_INET; /* Works because of the ns_c_in condition above */
				server->new_server = -1;
				server->current_weight = 0;
				server->weight = weight;
				server->fails = 0;
				server->accessed = -1;
				server->fail_probe = 0;
				server->max_fails = 1;
				server->fail_timeout = 10;

				ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "New server %V:%ud for %V", &server->uri.host, port, &domain->name);
			}
		}
	}

	if (ngx_array_init(&new_tokens, domain->state.tokens.pool, 10, sizeof(ngx_http_upstream_spdht_token_t)) != NGX_OK)
		return NGX_ERROR;

	/* For servers where the tokens are fresh, copy them to new_tokens */
	for (i = 0; i < domain->state.tokens.nelts; ++i) {
		ngx_http_upstream_spdht_token_t *old_token = ((ngx_http_upstream_spdht_token_t *) domain->state.tokens.elts) + i;

		if (old_token->server->new_server == -1) {
			ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Obsolete server %V:%ud for %V", old_token->server->uri.host, old_token->server->uri.port, &domain->name);
		} else if (old_token->server->tokens_tod >= ngx_time()) {
			ngx_http_upstream_spdht_server_t *new_server = ((ngx_http_upstream_spdht_server_t *) new_servers.elts) + old_token->server->new_server;
			ngx_http_upstream_spdht_token_t *new_token = (ngx_http_upstream_spdht_token_t *) ngx_array_push(&new_tokens);

			new_token->token.data = ngx_pstrdup(new_tokens.pool, &old_token->token);
			new_token->token.len = old_token->token.len;
			new_token->server = new_server;
		}
	}

	ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "Copied tokens: %ui (old %ui)", new_tokens.nelts, domain->state.tokens.nelts);

	for (i = 0; i < new_servers.nelts; ++i) {
		ngx_http_upstream_spdht_server_t *server = ((ngx_http_upstream_spdht_server_t *) new_servers.elts) + i;

		if (server->tokens_tod < ngx_time()) {
			/* Look up the tokens in "tokens.<port>.<host> TXT" */
			if (lookup_tokens(server, &new_tokens, log) != NGX_OK) {
				ngx_array_destroy(&new_tokens);
				ngx_array_destroy(&new_servers);
				return NGX_ERROR;
			}
		}

		// TODO: Also check port
		if (server->uri.host.len == server_name->len && !ngx_strncmp(server->uri.host.data, server_name->data, server->uri.host.len)) {
			domain->state.me = server;
			ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "Me in ring: %V", server_name);
		}
	}
	if (domain->state.me == NULL) {
		ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "Me not in ring: %V", server_name);
	}

	/* Sort the tokens to allow bisection */
	qsort(new_tokens.elts, new_tokens.nelts, new_tokens.size, ngx_http_upstream_spdht_token_cmp);

	ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "Copied tokens: %ui (%ui)", new_tokens.nelts, domain->state.tokens.nelts);

	/* Swap the servers, freeing the old array. */
	for (i = 0; i < domain->state.servers.nelts; ++i) {
		ngx_http_upstream_spdht_server_t *server = ((ngx_http_upstream_spdht_server_t *) domain->state.servers.elts) + i;

		if (server->new_server == -1) {
			ngx_pfree(domain->state.servers.pool, server->uri.host.data);
		}
	}

	ngx_array_destroy(&domain->state.servers);
	domain->state.servers = new_servers;

	/* Swap the tokens, freeing the old array. */
	for (i = 0; i < domain->state.tokens.nelts; ++i) {
		ngx_http_upstream_spdht_token_t *token = ((ngx_http_upstream_spdht_token_t *) domain->state.tokens.elts) + i;

		ngx_pfree(domain->state.tokens.pool, token->token.data);
	}

	ngx_array_destroy(&domain->state.tokens);
	domain->state.tokens = new_tokens;

	/* Parse config record */
	ngx_snprintf(txt_name, sizeof(txt_name), "config.%V.%V", srv_prefix, &domain->name)[0] = '\0';

	ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "res_query(txt, %s)", txt_name);

	anslen = res_query((const char*) txt_name, ns_c_in, ns_t_txt, ans, sizeof(ans));

	if (anslen < 0) {
		ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to look up TXT RR: %d (TXT record for %s)", h_errno, txt_name);
		return NGX_ERROR;
	}

	if (ns_initparse(ans, anslen, &msg)) {
		ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse TXT answer: %d (TXT record for %s)", h_errno, txt_name);
		return NGX_ERROR;
	}

	nans = ns_msg_count(msg, ns_s_an);

	for (i = 0; i < nans; ++i) {
		ns_rr rr;

		if (ns_parserr(&msg, ns_s_an, i, &rr)) {
			ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse RR in TXT answer: %d", h_errno);
			return NGX_ERROR;
		}

		if (ns_rr_type(rr) != ns_t_txt || ns_rr_class(rr) != ns_c_in)
			continue;

		if (ns_rr_ttl(rr) < min_ttl)
			min_ttl = ns_rr_ttl(rr);

		{
			const u_char *data = ns_rr_rdata(rr);
			ngx_uint_t remaining = ns_rr_rdlen(rr);

			while (remaining) {
				if (data[0] > remaining - 1) {
					ngx_log_error(NGX_LOG_ERR, log, 0, "Failed to parse TXT rdata");
					return NGX_ERROR;
				}

				ngx_str_t conf_value = { len: data[0], data: (u_char*) data + 1 };

				if (ngx_strncmp(conf_value.data, "slaves=", 7) == 0) {
					ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "TXT slaves %V", &conf_value);
					ngx_str_t slaves = { len: data[0] - 7, data: (u_char*) data + 8 };
					parse_slaves(domain, slaves, log);
				}

				remaining -= data[0] + 1;
				data += data[0] + 1;
			}
		}

	}

	if (min_ttl != UINT_MAX)
		domain->state.tod = ngx_time() + min_ttl;

	ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "SRV %s -> %ud hosts, TTL %ud", srv_name, nans, min_ttl);

	ngx_http_upstream_spdht_domain_dump(domain, log);

	return NGX_OK;
}

ngx_int_t ngx_http_upstream_spdht_validate_server(ngx_http_upstream_spdht_server_t *server, ngx_pool_t *pool, ngx_log_t *log)
{
	if (ngx_inet_resolve_host(pool, &server->uri) != NGX_OK)
		return NGX_ERROR;

	return NGX_OK;
}
