/*
 * Copyright © 2006-2009 Spotify AB
 */

/* Must be at the top to have feature macros set */
#include <ngx_config.h>
#include "ngx_http_upstream_spdht_module.h"


/** Get a server from a ring. */
static ngx_http_upstream_spdht_server_t *
token_server(ngx_http_upstream_spdht_domain_t *domain, ngx_int_t node)
{
	ngx_uint_t nelts = domain->state.tokens.nelts;
	if (nelts == 0) {
		return NULL;
	}
	ngx_http_upstream_spdht_token_t *elts = (ngx_http_upstream_spdht_token_t *) domain->state.tokens.elts;
	return elts[node % nelts].server;
}


/** Get the next round_robin weighted server from a ring segment. */
static ngx_http_upstream_spdht_server_t *
weight_server(ngx_http_upstream_spdht_domain_t *domain, ngx_array_t *servers /* (ngx_http_upstream_spdht_server_t *) */, ngx_log_t *log)
{
	ngx_int_t nelts = domain->state.tokens.nelts;
	ngx_int_t zero_weights = 0;
	ngx_int_t n = servers->nelts;
	ngx_http_upstream_spdht_server_t **selts = (ngx_http_upstream_spdht_server_t **) servers->elts;

	ngx_int_t t, j;

	for (t = 0; t < 2; ++t) {
		/* Check the ring segment given by servers */
		ngx_http_upstream_spdht_server_t *server = NULL;
		for (j = 0; j < n; ++j) {
			ngx_http_upstream_spdht_server_t *candidate = selts[j];
			if (domain->state.me != NULL) {
				ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: check me %V:%ud", &domain->state.me->uri.host, domain->state.me->uri.port);
			}
			if (candidate == NULL || candidate == domain->state.me) {
				continue;
			}

			/* Only consider servers with positive current weight and weight */
			if (candidate->current_weight <= 0 || candidate->weight <= 0) {
				ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: zero weight %V:%ud", &candidate->uri.host, candidate->uri.port);
				++zero_weights;
				continue;
			}

			/* Select the server with highest current weight / weight */
			if (server == NULL ||
					1000 * candidate->current_weight / candidate->weight >
					1000 * server->current_weight / server->weight) {
				server = candidate;
			}
		}

		if (server != NULL) {
			/* Decrement current weight */
			--server->current_weight;
			return server;
		}

		/* No valid servers found */
		if (zero_weights == 0) {
			ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: no valid servers");
			break;
		}

		/* Current weights in this ring segment are zero */
		/* Reset _all_ current weights in the whole ring */
		ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: reset weights %V", &domain->name);
		time_t now = ngx_time();
		for (j = 0; j < nelts; ++j) {
			ngx_http_upstream_spdht_server_t *reset = token_server(domain, j);
			if (reset != NULL) {
				ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: reset %V:%ud to %i", &reset->uri.host, reset->uri.port, reset->weight);
				reset->current_weight = reset->weight;
				/* Clear fail probe flag after fail timeout */
				if (now - reset->accessed > reset->fail_timeout) {
					reset->fail_probe = 0;
				}
			}
		}
	}
	return NULL;
}


/** Get a round robin and failure weighted server from a ring segment */
ngx_http_upstream_spdht_server_t *
ngx_http_upstream_spdht_failover_server(ngx_http_upstream_spdht_domain_t *domain, ngx_array_t *servers /* (ngx_http_upstream_spdht_server_t *) */, ngx_log_t *log)
{
	ngx_int_t nelts = domain->state.tokens.nelts;
	time_t now = ngx_time();
	ngx_int_t n = servers->nelts;

	ngx_int_t t, j;

	for (t = 0; t < 2; ++t) {
		for (j = 0; j < n; ++j) {
			ngx_http_upstream_spdht_server_t *server = weight_server(domain, servers, log);
			if (server == NULL) {
				return NULL;
			}
			/* If probe is set, decrease current weight by one half failure */
			if (server->max_fails > 0 && server->fail_probe > 0) {
				server->current_weight -= server->weight / server->max_fails / 2;
			}
			/* Return server if it is within max fails */
			if (server->max_fails == 0 || server->fails < server->max_fails) {
				ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: server %V:%ud weight %i", &server->uri.host, server->uri.port, server->current_weight);
				return server;
			}
			/* Clear and return server after fail timeout */
			if (now - server->accessed > server->fail_timeout) {
				ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: fail timeout %V:%ud weight %i", &server->uri.host, server->uri.port, server->current_weight);
				server->fails = 0;
				server->fail_probe = 1; /* but set fail probe flag */
				server->accessed = now;
				return server;
			}
			/* Clear current weight and try next */
			server->current_weight = 0;
		}

		/* No live server found in ring segment */
		/* If there are live servers in the whole ring, fail */
		for (j = 0; j < nelts; ++j) {
			ngx_http_upstream_spdht_server_t *server = token_server(domain, j);
			if (server->max_fails == 0 || server->fails < server->max_fails) {
				if (server->fail_probe == 0) {
					ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: live but not in segment");
					return NULL;
				}
			}
		}

		/* If all ring servers fail, mark them as live for quick recovery */
		ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: quick recovery %V", &domain->name);
		for (j = 0; j < nelts; ++j) {
			ngx_http_upstream_spdht_server_t *server = token_server(domain, j);
			server->fails = 0;
			server->fail_probe = 0;
		}
	}
	ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "Failover: second fail %V", &domain->name);
	return NULL;
}


/** Mark a server failure */
void
ngx_http_upstream_spdht_fail_server(ngx_http_upstream_spdht_server_t *server, ngx_log_t *log)
{
	time_t now = ngx_time();

	++server->fails;
	server->accessed = now;

	/* Decrease current weight by one failure */
	if (server->max_fails > 0) {
		server->current_weight -= server->weight / server->max_fails;
		if (server->current_weight < 0) {
			server->current_weight = 0;
		}
	}
}
