/*
 * Copyright © 2006-2009 Spotify AB
 */

#include "ngx_http_upstream_spdht_module.h"



/* --- Types --- */
typedef struct {
	ngx_http_upstream_spdht_srv_conf_t *usdcf;
	ngx_http_request_t *request;
	ngx_str_t *server_name;
	ngx_str_t hash;
	ngx_uint_t state;
	ngx_http_upstream_spdht_domain_t *domain;
	ngx_array_t servers;
	ngx_uint_t node;
	ngx_uint_t rem_nodes;
	ngx_int_t num_tries;
	ngx_http_upstream_spdht_server_t *current_server;

	ngx_http_upstream_spdht_server_t *data_origin;
	ngx_int_t next_dht_level;
} ngx_http_upstream_spdht_peer_data_t;


/* --- Functions --- */
static char * ngx_http_upstream_spdht(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static void * ngx_http_upstream_spdht_create_loc_conf(ngx_conf_t *cf);
static char * ngx_http_upstream_spdht_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child);
static char * ngx_http_upstream_spdht_max_hops(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static char * ngx_http_upstream_spdht_via_slaves_flag(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static char * ngx_http_upstream_spdht_purge_handler_flag(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static char * ngx_http_upstream_spdht_put_handler_flag(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static ngx_int_t ngx_http_upstream_spdht_init(ngx_conf_t *cf, ngx_http_upstream_srv_conf_t *us);
static ngx_int_t ngx_http_upstream_spdht_init_peer(ngx_http_request_t *r, ngx_http_upstream_srv_conf_t *us);
static ngx_int_t ngx_http_upstream_spdht_peer_get(ngx_peer_connection_t *pc, void *data);
static void ngx_http_upstream_spdht_free_peer(ngx_peer_connection_t *pc, void *data, ngx_uint_t state);


/* --- Data --- */
static ngx_command_t  ngx_http_upstream_spdht_commands[] = {

	{ ngx_string("spdht"),
	  NGX_HTTP_UPS_CONF|NGX_CONF_2MORE, /* spdht <srv-prefix> <my-domain> [<parent-domain>...] */
	  ngx_http_upstream_spdht,
	  0,
	  0,
	  NULL },

	{ ngx_string("spdht_max_hops"),
	  NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1, /* spdht_max_hops n */
	  ngx_http_upstream_spdht_max_hops,
	  NGX_HTTP_LOC_CONF_OFFSET,
	  0,
	  NULL },

	{ ngx_string("spdht_via_slaves"),
	  NGX_HTTP_LOC_CONF|NGX_CONF_NOARGS, /* spdht_via_slaves {on|off} */
	  ngx_http_upstream_spdht_via_slaves_flag,
	  NGX_HTTP_LOC_CONF_OFFSET,
	  0,
	  NULL },

	{ ngx_string("spdht_purge_handler"),
	  NGX_HTTP_LOC_CONF|NGX_CONF_NOARGS, /* spdht_purge_handler {on|off} */
	  ngx_http_upstream_spdht_purge_handler_flag,
	  NGX_HTTP_LOC_CONF_OFFSET,
	  0,
	  NULL },

	{ ngx_string("spdht_put_handler"),
	  NGX_HTTP_LOC_CONF|NGX_CONF_NOARGS, /* spdht_put_handler {on|off} */
	  ngx_http_upstream_spdht_put_handler_flag,
	  NGX_HTTP_LOC_CONF_OFFSET,
	  0,
	  NULL },

	  ngx_null_command
};

static ngx_http_module_t  ngx_http_upstream_spdht_module_ctx = {
	NULL,								  /* preconfiguration */
	NULL,								  /* postconfiguration */

	NULL,								  /* create main configuration */
	NULL,								  /* init main configuration */

	NULL,								  /* create server configuration */
	NULL,								  /* merge server configuration */

	ngx_http_upstream_spdht_create_loc_conf,			  /* create location configuration */
	ngx_http_upstream_spdht_merge_loc_conf				  /* merge location configuration */
};

ngx_module_t  ngx_http_upstream_spdht_module = {
	NGX_MODULE_V1,
	&ngx_http_upstream_spdht_module_ctx, /* module context */
	ngx_http_upstream_spdht_commands,	/* module directives */
	NGX_HTTP_MODULE,					   /* module type */
	NULL,								  /* init master */
	NULL,								  /* init module */
	NULL,								  /* init process */
	NULL,								  /* init thread */
	NULL,								  /* exit thread */
	NULL,								  /* exit process */
	NULL,								  /* exit master */
	NGX_MODULE_V1_PADDING
};


static void * ngx_http_upstream_spdht_create_loc_conf(ngx_conf_t *cf)
{
	ngx_http_upstream_spdht_loc_conf_t  *conf = ngx_pcalloc(cf->pool, sizeof(ngx_http_upstream_spdht_loc_conf_t));
	if (conf == NULL) {
		return NGX_CONF_ERROR;
	}

	conf->max_hops = 0;
	conf->via_slaves = 0;
	conf->handle_purge = 0;
	conf->handle_put = 0;

	return conf;
}

static char * ngx_http_upstream_spdht_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
{
	ngx_http_upstream_spdht_loc_conf_t *prev = parent;
	ngx_http_upstream_spdht_loc_conf_t *conf = child;

	if (!conf->max_hops)
		conf->max_hops = prev->max_hops;

	if (!conf->via_slaves)
		conf->via_slaves = prev->via_slaves;

	if (!conf->handle_purge)
		conf->handle_purge = prev->handle_purge;

	if (!conf->handle_put)
		conf->handle_put = prev->handle_put;

	return NGX_CONF_OK;
}

static char * ngx_http_upstream_spdht(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
	ngx_http_upstream_srv_conf_t  *uscf = ngx_http_conf_get_module_srv_conf(cf, ngx_http_upstream_module);
	ngx_http_upstream_spdht_srv_conf_t *udscf = ngx_pcalloc(cf->pool, sizeof(ngx_http_upstream_spdht_srv_conf_t));
	ngx_uint_t i;
	ngx_str_t *values = (ngx_str_t *) cf->args->elts + 1;

	if (udscf == NULL) {
		return NGX_CONF_ERROR;
	}

	udscf->srv_prefix.data = ngx_pstrdup(cf->pool, &values[0]);
	if (udscf->srv_prefix.data == NULL) {
		return NGX_CONF_ERROR;
	}

	udscf->srv_prefix.len = values[0].len;

	udscf->domain.name.data = ngx_pstrdup(cf->pool, &values[1]);
	if (udscf->domain.name.data == NULL) {
		return NGX_CONF_ERROR;
	}

	udscf->domain.name.len = values[1].len;

	udscf->domain.state.tod = -1;
	udscf->domain.state.num_slaves = 0;
	ngx_memset(&udscf->domain.state.servers, 0, sizeof(udscf->domain.state.servers));
	ngx_memset(&udscf->domain.state.tokens, 0, sizeof(udscf->domain.state.tokens));
	ngx_memset(&udscf->domain.state.collections, 0, sizeof(udscf->domain.state.collections));
	udscf->domain.state.me = NULL;

	/* These are used in validate_domain_state(), but we don't need the allocation yet */
	udscf->domain.state.servers.pool = cf->pool;
	udscf->domain.state.tokens.pool = cf->pool;

	if (ngx_array_init(&udscf->parents, cf->pool, 3, sizeof(ngx_http_upstream_spdht_domain_t)) != NGX_OK) {
		return NGX_CONF_ERROR;
	}

	for (i = 2; i < cf->args->nelts - 1; i++) {
		ngx_str_t *v = &values[i];
		ngx_http_upstream_spdht_domain_t *d = (ngx_http_upstream_spdht_domain_t *) ngx_array_push(&udscf->parents);

		d->name.data = ngx_pstrdup(cf->pool, v);
		if (d->name.data == NULL) {
			return NGX_CONF_ERROR;
		}

		d->name.len = v->len;

		d->state.tod = -1;
		d->state.num_slaves = 0;
		ngx_memset(&d->state.servers, 0, sizeof(d->state.servers));
		ngx_memset(&d->state.tokens, 0, sizeof(d->state.tokens));
		ngx_memset(&d->state.collections, 0, sizeof(d->state.collections));
		d->state.me = NULL;

		/* These are used in validate_domain_state(), but we don't need the allocation yet */
		d->state.servers.pool = cf->pool;
		d->state.tokens.pool = cf->pool;
	}

	ngx_http_conf_upstream_srv_conf(uscf, ngx_http_upstream_spdht_module) = udscf;
	uscf->peer.init_upstream = ngx_http_upstream_spdht_init;
	uscf->flags = 0;

	return ngx_http_upstream_spdht_dns_init(cf);
}

static ngx_int_t ngx_http_upstream_spdht_purge_handler(ngx_http_request_t *r)
{
	ngx_log_error(NGX_LOG_EMERG, r->pool->log, 0, "DO PURGE %V", &r->uri);
	return NGX_HTTP_INTERNAL_SERVER_ERROR;
}

static ngx_int_t ngx_http_upstream_spdht_put_handler(ngx_http_request_t *r)
{
	ngx_log_error(NGX_LOG_EMERG, r->pool->log, 0, "DO PUT %V", &r->uri);
	return NGX_HTTP_INTERNAL_SERVER_ERROR;
}

static ngx_int_t ngx_http_upstream_spdht_handler(ngx_http_request_t *r)
{
	static const ngx_str_t PURGE = ngx_string("PURGE");
	ngx_http_upstream_spdht_loc_conf_t *slcf = (ngx_http_upstream_spdht_loc_conf_t *) ngx_http_get_module_loc_conf(r, ngx_http_upstream_spdht_module);

	if (!slcf->orig_handler) {
		ngx_log_error(NGX_LOG_EMERG, r->pool->log, 0, "No original handler!");
		return NGX_ERROR;
	}

	if (slcf->handle_purge && r->method == NGX_HTTP_UNKNOWN && r->method_name.len == PURGE.len && !ngx_strncasecmp(r->method_name.data, PURGE.data, PURGE.len)) {
		return ngx_http_upstream_spdht_purge_handler(r);
	}

	if (slcf->handle_put && r->method == NGX_HTTP_PUT) {
		return ngx_http_upstream_spdht_put_handler(r);
	}

	return slcf->orig_handler(r);
}

static char * ngx_http_upstream_spdht_max_hops(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
	ngx_http_upstream_spdht_loc_conf_t *slcf = (ngx_http_upstream_spdht_loc_conf_t *) conf;
	ngx_str_t *values = (ngx_str_t *) cf->args->elts + 1;

	slcf->max_hops = ngx_atoi(values[0].data, values[0].len);

	return NGX_CONF_OK;
}

static char * ngx_http_upstream_spdht_via_slaves_flag(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
	ngx_http_upstream_spdht_loc_conf_t *slcf = (ngx_http_upstream_spdht_loc_conf_t *) conf;

	slcf->via_slaves = 1;

	return NGX_CONF_OK;
}

static char * ngx_http_upstream_spdht_purge_handler_flag(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
	ngx_http_core_loc_conf_t *clcf = (ngx_http_core_loc_conf_t *) ngx_http_conf_get_module_loc_conf(cf, ngx_http_core_module);
	ngx_http_upstream_spdht_loc_conf_t *slcf = (ngx_http_upstream_spdht_loc_conf_t *) conf;

	slcf->handle_purge = 1;

	if (!slcf->orig_handler) {
		slcf->orig_handler = clcf->handler;
		clcf->handler = ngx_http_upstream_spdht_handler;
	}

	return NGX_CONF_OK;
}

static char * ngx_http_upstream_spdht_put_handler_flag(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
	ngx_http_core_loc_conf_t *clcf = (ngx_http_core_loc_conf_t *) ngx_http_conf_get_module_loc_conf(cf, ngx_http_core_module);
	ngx_http_upstream_spdht_loc_conf_t *slcf = (ngx_http_upstream_spdht_loc_conf_t *) conf;

	slcf->handle_put = 1;

	if (!slcf->orig_handler) {
		slcf->orig_handler = clcf->handler;
		clcf->handler = ngx_http_upstream_spdht_handler;
	}

	return NGX_CONF_OK;
}

static ngx_int_t ngx_http_upstream_spdht_init(ngx_conf_t *cf, ngx_http_upstream_srv_conf_t *us)
{
	us->peer.init = ngx_http_upstream_spdht_init_peer;

	return NGX_OK;
}

static ngx_int_t num_slaves(ngx_http_request_t *r, ngx_http_upstream_spdht_domain_t *domain) {
	/* Search for /.../collection/... in uri */
	ngx_uint_t count = 0, start = 0, off, i;
	for (off = 0; off < r->uri.len; ++off) {
		if (r->uri.data[off] == '/') {
			++count;
			if (count == 1)
				start = off + 1;
			else if (count == 2) {
				ngx_str_t collection = { len: off - start, data: r->uri.data + start };
				for (i = 0; i < domain->state.collections.nelts; ++i) {
					ngx_http_upstream_spdht_collection_t *c = ((ngx_http_upstream_spdht_collection_t *) domain->state.collections.elts) + i;
					if (c->name.len == collection.len &&
							ngx_strncmp(c->name.data, collection.data, collection.len) == 0) {
						ngx_log_debug3(NGX_LOG_DEBUG_HTTP, r->upstream->peer.log, 0, "num_slaves %V %V:%i", &domain->name, &c->name, c->num_slaves);
						return c->num_slaves;
					}
				}
			}
		}
	}
	ngx_log_debug3(NGX_LOG_DEBUG_HTTP, r->upstream->peer.log, 0, "num_slaves %V :%i", &domain->name, domain->state.num_slaves);
	return domain->state.num_slaves;
}


static const ngx_str_t X_SPOTIFY_DATA_ORIGIN = ngx_string("x-spotify-data-origin");
static const ngx_str_t X_SPOTIFY_DHT_DOMAIN = ngx_string("x-spotify-dht-domain");
static const ngx_str_t X_FORWARDED_FOR = ngx_string("x-forwarded-for");


static ngx_int_t ngx_http_upstream_spdht_init_peer(ngx_http_request_t *r, ngx_http_upstream_srv_conf_t *uscf)
{
	ngx_http_core_srv_conf_t *hccf = (ngx_http_core_srv_conf_t *) ngx_http_get_module_srv_conf(r, ngx_http_core_module);
	ngx_http_upstream_spdht_srv_conf_t *usdcf = (ngx_http_upstream_spdht_srv_conf_t *) ngx_http_conf_upstream_srv_conf(uscf, ngx_http_upstream_spdht_module);
	ngx_http_upstream_spdht_peer_data_t *usdpd;
	ngx_http_upstream_spdht_loc_conf_t *slcf = (ngx_http_upstream_spdht_loc_conf_t *) ngx_http_get_module_loc_conf(r, ngx_http_upstream_spdht_module);

	usdpd = (ngx_http_upstream_spdht_peer_data_t *) r->upstream->peer.data;
	if (usdpd == NULL) {
		usdpd = (ngx_http_upstream_spdht_peer_data_t *) ngx_pcalloc(r->pool, sizeof(*usdpd));
		if (usdpd == NULL) {
			return NGX_ERROR;
		}

		r->upstream->peer.data = usdpd;
	}

	r->upstream->peer.get = ngx_http_upstream_spdht_peer_get;
	r->upstream->peer.free = ngx_http_upstream_spdht_free_peer;
	usdpd->usdcf = usdcf;
	usdpd->request = r;
	usdpd->server_name = &hccf->server_name;

	if (ngx_http_upstream_spdht_validate_domain_state(&usdcf->srv_prefix, &usdcf->domain, &hccf->server_name, r->upstream->peer.log) != NGX_OK) {
		return NGX_ERROR;
	}

	/* Master, slaves, and next domain (will update tries) */
	r->upstream->peer.tries = 1 + num_slaves(r, &usdcf->domain) + 1;

	/* Parse the module specific headers */
	{
		ngx_list_part_t *part = &r->headers_in.headers.part;
		ngx_table_elt_t *data = (ngx_table_elt_t *) part->elts;
		ngx_uint_t i;

		for (i = 0; i < part->nelts || part->next; ++i) {
			if (i == part->nelts) {
				part = part->next;
				data = (ngx_table_elt_t *) part->elts;
				i = 0;
			}

			if (data[i].key.len == X_SPOTIFY_DATA_ORIGIN.len && !ngx_strncmp(data[i].lowcase_key, X_SPOTIFY_DATA_ORIGIN.data, X_SPOTIFY_DATA_ORIGIN.len)) {
				usdpd->data_origin = (ngx_http_upstream_spdht_server_t *) ngx_pcalloc(r->pool, sizeof(ngx_http_upstream_spdht_server_t));
				usdpd->data_origin->tokens_tod = -1;
				usdpd->data_origin->uri.url.len = data[i].value.len;
				usdpd->data_origin->uri.url.data = ngx_pstrdup(r->pool, &data[i].value);
				usdpd->data_origin->new_server = -1;
			}
			if (data[i].key.len == X_SPOTIFY_DHT_DOMAIN.len && !ngx_strncmp(data[i].lowcase_key, X_SPOTIFY_DHT_DOMAIN.data, X_SPOTIFY_DHT_DOMAIN.len)) {
				if (data[i].value.len == usdcf->domain.name.len && !ngx_strncmp(data[i].value.data, usdcf->domain.name.data, usdcf->domain.name.len)) {
					usdpd->next_dht_level = 1;
					ngx_log_debug2(NGX_LOG_DEBUG_HTTP, r->upstream->peer.log, 0, "next_dht_level set %V", &usdcf->domain.name);
				}
			}
			if (data[i].key.len == X_FORWARDED_FOR.len && !ngx_strncmp(data[i].lowcase_key, X_FORWARDED_FOR.data, X_FORWARDED_FOR.len)) {
				ngx_uint_t hops = 1, j;
				for (j = 0; j < data[i].value.len; ++j) {
					if (data[i].value.data[j] == ',')
						++hops;
				}
				ngx_log_debug3(NGX_LOG_DEBUG_HTTP, r->upstream->peer.log, 0, "X-Forwarded-For %V has %d hops, maximum is %d", &data[i].value, hops, slcf->max_hops);
				if (slcf->max_hops && hops > slcf->max_hops) {
					ngx_log_error(NGX_LOG_WARN, r->upstream->peer.log, 0, "X-Forwarded-Fox too many hops (%d > %d) for URI %V", hops, slcf->max_hops, &r->unparsed_uri);
					return NGX_ERROR;
				}
			}
		}
	}

	if (ngx_http_upstream_spdht_calculate_hash(&usdpd->hash, r) != NGX_OK) {
		ngx_log_error(NGX_LOG_ERR, r->upstream->peer.log, 0, "Failed to calculate hash for URI %V", &r->unparsed_uri);
		return NGX_ERROR;
	}

	{
		u_char buf[256];

		*ngx_hex_dump(buf, usdpd->hash.data, usdpd->hash.len) = 0;
		ngx_log_debug3(NGX_LOG_DEBUG_HTTP, r->upstream->peer.log, 0, "INIT %ui %ui %s %V", usdpd->node, r->upstream->peer.tries, buf, &r->uri);
	}

	return NGX_OK;
}

/*
 * The attempt order is:
 *
 *  1.  The domain master
 *      if we are a slave and the master seems to be up
 *      if the X-Spotify-Dht-Domain header was given
 *      and equal to the local domain, skip this step
 *
 *  2.  The data origin
 *      if the X-Spotify-Data-Origin header was given
 *      TODO: we should add security to this...
 *
 *  3.  The master (or slave) of the parent domain
 *      if master or some slaves seem to be up
 *
 *  4.  The master (or slave) of the next domain
 *      if master or some slaves seem to be up
 *
 */
static ngx_int_t ngx_http_upstream_spdht_peer_get(ngx_peer_connection_t *pc, void *data)
{
	ngx_http_upstream_spdht_peer_data_t *usdpd = (ngx_http_upstream_spdht_peer_data_t *) data;
	ngx_http_upstream_spdht_loc_conf_t *slcf = (ngx_http_upstream_spdht_loc_conf_t *) ngx_http_get_module_loc_conf(usdpd->request, ngx_http_upstream_spdht_module);
	ngx_http_upstream_spdht_server_t *server = NULL;

	ngx_log_debug3(NGX_LOG_DEBUG_HTTP, pc->log, 0, "GET state(%ui) domain(%p) node(%ui) rem_nodes(%ui) num_tries(%i)", usdpd->state, usdpd->domain, usdpd->node, usdpd->rem_nodes, usdpd->num_tries);

	switch (usdpd->state) {
	case 0:
		++usdpd->state;
		usdpd->domain = &usdpd->usdcf->domain;
		//usdpd->node = ngx_http_upstream_spdht_bisect_left(&usdpd->hash, usdpd->domain->state.tokens.elts, usdpd->domain->state.tokens.nelts, sizeof(ngx_http_upstream_spdht_token_t), ngx_http_upstream_spdht_token_cmp);
		if (ngx_array_init(&usdpd->servers, usdpd->request->pool, 1, sizeof(ngx_http_upstream_spdht_server_t *)) != NGX_OK) {
			return NGX_ERROR;
		}
		ngx_http_upstream_spdht_master_slaves(usdpd->domain, usdpd->hash, slcf->via_slaves ? 1 + num_slaves(usdpd->request, usdpd->domain) : 1, &usdpd->servers, pc->log);

		/* Domain master */
		server = ngx_http_upstream_spdht_failover_server(usdpd->domain, &usdpd->servers, pc->log);

		if (server != NULL && !usdpd->next_dht_level) {
			ngx_log_debug3(NGX_LOG_DEBUG_HTTP, pc->log, 0, "Trying sibling master server %V:%ud", &server->uri.host, server->uri.port);
			break;
		}

		/* drop through */
	case 1:
		++usdpd->state;

		/* Data origin */
		server = usdpd->data_origin;
		if (server != NULL) {
			ngx_log_debug3(NGX_LOG_DEBUG_HTTP, pc->log, 0, "Trying origin server %V:%ud", &server->uri.host, server->uri.port);
			break;
		}

		/* drop through */
	case 2:
		++usdpd->state;

		/* Parent domain */
		usdpd->domain = (ngx_http_upstream_spdht_domain_t *) usdpd->usdcf->parents.elts;
		if (usdpd->domain >= ((ngx_http_upstream_spdht_domain_t *) usdpd->usdcf->parents.elts) + usdpd->usdcf->parents.nelts) {
			return NGX_BUSY;
		}

		usdpd->rem_nodes = num_slaves(usdpd->request, usdpd->domain);
		usdpd->num_tries = 0;
		usdpd->servers.nelts = 0;

		/* drop through */
	case 3:
		do {
			while (ngx_http_upstream_spdht_validate_domain_state(&usdpd->usdcf->srv_prefix, usdpd->domain, usdpd->server_name, pc->log) != NGX_OK ||
						 usdpd->num_tries >= 1 + num_slaves(usdpd->request, usdpd->domain)) {
				/* Next domain */
				++usdpd->domain;
				if (usdpd->domain >= ((ngx_http_upstream_spdht_domain_t *) usdpd->usdcf->parents.elts) + usdpd->usdcf->parents.nelts) {
					return NGX_BUSY;
				}
				usdpd->rem_nodes = num_slaves(usdpd->request, usdpd->domain);
				usdpd->num_tries = 0;
				usdpd->servers.nelts = 0;
				/* Master, slaves, and next domain (will update tries again) */
				pc->tries = 1 + num_slaves(usdpd->request, usdpd->domain) + 1;
			}

			//usdpd->node = ngx_http_upstream_spdht_bisect_left(&usdpd->hash, usdpd->domain->state.tokens.elts, usdpd->domain->state.tokens.nelts, sizeof(ngx_http_upstream_spdht_token_t), ngx_http_upstream_spdht_token_cmp);
			if (usdpd->servers.nelts == 0) {
				/* Update servers */
				if (ngx_http_upstream_spdht_master_slaves(usdpd->domain, usdpd->hash, 1 + num_slaves(usdpd->request, usdpd->domain), &usdpd->servers, pc->log) != NGX_OK) {
					return NGX_ERROR;
				}
			}

			/* Next server */
			server = ngx_http_upstream_spdht_failover_server(usdpd->domain, &usdpd->servers, pc->log);

			++usdpd->num_tries;
		} while (server == NULL);

		ngx_log_debug3(NGX_LOG_DEBUG_HTTP, pc->log, 0, "Trying parent server %V:%ud (domain %V)", &server->uri.host, server->uri.port, &usdpd->domain->name);
		break;

	default:
		return NGX_BUSY;
	}

	if (ngx_http_upstream_spdht_validate_server(server, usdpd->request->pool, pc->log) != NGX_OK)
		return NGX_ERROR;

	if (server->uri.naddrs < 1)
		return NGX_BUSY;

	usdpd->current_server = server;

	pc->name = &server->uri.addrs[0].name;
	pc->sockaddr = (struct sockaddr *) server->uri.addrs[0].sockaddr;
	pc->socklen = server->uri.addrs[0].socklen;

	return NGX_OK;
}

static void ngx_http_upstream_spdht_free_peer(ngx_peer_connection_t *pc, void *data, ngx_uint_t state)
{
	ngx_http_upstream_spdht_peer_data_t *usdpd = (ngx_http_upstream_spdht_peer_data_t *) data;

	if (state & NGX_PEER_FAILED) {
		ngx_http_upstream_spdht_fail_server(usdpd->current_server, pc->log);
	}

	if (pc->tries) {
		--pc->tries;
	}
}
