/*
 * Copyright © 2006-2009 Spotify AB
 */

/* Must be at the top to have feature macros set */
#include <ngx_config.h>
#include <ngx_sha1.h>
#include "ngx_http_upstream_spdht_module.h"



static ngx_int_t parse_digit(ngx_int_t c)
{
	if (c <  '0') return -1;
	if (c <= '9') return c - '0';
	if (c <  'A') return -1;
	if (c <= 'Z') return c - 'A' + 10;
	if (c <  'a') return -1;
	if (c <= 'z') return c - 'a' + 10;

	return -1;
}

ngx_int_t ngx_http_upstream_spdht_bisect_left(const void *key, const void *_base, size_t nmemb, size_t size, int (*compar)(const void *, const void *))
{
	const u_char *base = (const u_char *) _base;
	int a = 0, b = nmemb;

	while (a < b) {
		int m = (a + b) / 2;
		int r = compar(key, base + m * size);

		if (r > 0)
			a = m + 1;
		else
			b = m;
	}

	return b;
}

ngx_int_t ngx_http_upstream_spdht_calculate_hash(ngx_str_t *dest, const ngx_http_request_t *r)
{
	dest->data = (u_char *) ngx_pcalloc(r->pool, 20);
	ngx_sha1_t ctx;

	if (!dest->data)
		return NGX_ERROR;

	if (!ngx_sha1_init(&ctx) || !ngx_sha1_update(&ctx, r->uri.data, r->uri.len) || !ngx_sha1_final((unsigned char *) dest->data, &ctx))
		return NGX_ERROR;

	dest->len = 20;

	return NGX_OK;
}

void ngx_http_upstream_spdht_domain_dump(const ngx_http_upstream_spdht_domain_t *d, ngx_log_t *log)
{
	unsigned int i;

	ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "--- Domain %V ---", &d->name);
	ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "  SRV TTL: %d", (d->state.tod >= 0 ? d->state.tod - ngx_time() : -1));
	ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "  Slaves : %i", d->state.num_slaves);

	for (i = 0; i < d->state.servers.nelts; ++i) {
		ngx_http_upstream_spdht_server_t *s = ((ngx_http_upstream_spdht_server_t *) d->state.servers.elts) + i;

		ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "---- Server %ud %V:%ud ----", i, &s->uri.host, s->uri.port);
		ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "  Tokens TTL: %d", (s->tokens_tod >= 0 ? s->tokens_tod - ngx_time() : -1));
		ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "  New       : %i", s->new_server);
	}

	ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "---- Tokens %ui ----", d->state.tokens.nelts);

	for (i = 0; i < d->state.tokens.nelts; ++i) {
		ngx_http_upstream_spdht_token_t *t = ((ngx_http_upstream_spdht_token_t *) d->state.tokens.elts) + i;
		u_char buf[256];

		ngx_hex_dump(buf, t->token.data, t->token.len)[0] = '\0';
		ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "%ud %s %V:%ud", i, buf, &t->server->uri.host, t->server->uri.port);
	}

	ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "---- Collections %ui ----", d->state.collections.nelts);

	for (i = 0; i < d->state.collections.nelts; ++i) {
		ngx_http_upstream_spdht_collection_t *c = ((ngx_http_upstream_spdht_collection_t *) d->state.collections.elts) + i;
		ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0, "%ud %V:%i", i, &c->name, c->num_slaves);
	}
}

ngx_int_t ngx_http_upstream_spdht_parse_hex_string(ngx_pool_t *pool, ngx_str_t *dest, const ngx_str_t *src)
{
	size_t i;

	if (src->len & 1)
		return NGX_ERROR;

	dest->data = ngx_pcalloc(pool, src->len / 2);
	dest->len = src->len / 2;

	if (!dest->data)
		return NGX_ERROR;

	for (i = 0; i < dest->len; ++i) {
		dest->data[i] =
			(parse_digit(src->data[i * 2]) << 4) |
			parse_digit(src->data[i * 2 + 1]);
	}

	return NGX_OK;
}

int ngx_http_upstream_spdht_token_cmp(const void *a, const void *b)
{
	const ngx_http_upstream_spdht_token_t *aa = (ngx_http_upstream_spdht_token_t *) a;
	const ngx_http_upstream_spdht_token_t *bb = (ngx_http_upstream_spdht_token_t *) b;
	int prefix = ngx_memcmp(aa->token.data, bb->token.data, (aa->token.len < bb->token.len ? aa->token.len : bb->token.len));

	if (prefix)
		return prefix;

	return aa->token.len - bb->token.len;
}
